#!/usr/bin/perl

BEGIN {
if ($ARGV[0] eq "--help"){
   print "Delete entries from an LDAP server\n\n",
         "Usage:  delete.pl [options] [dn1 [dn2 ..]]\n",
         "         dn1 dn2 .. DNs to delete. If not given it will be readed sequentional\n",
         "                    from stdin or from the file specified with \"-f file\" \n",
         "                    in format 'dn: <dn>'\n",   
         "Options:\n",
         "\t -i            prompt for confirmation before deleting any entry\n",             
         "\t -f file       read operations from \'file\'\n",
         "\t -c            continue deleting next entries after reporting an error\n",
         "\t -y            delete all subtree without confirmation\n",
         "\t -h host       LDAP serwer (default: localhost)\n",
         "\t -p port       port on LDAP server (default: 389)\n",
         "\t -v version    protocol version to be used (default: 3)\n",
         "\t -d binddn     bind DN\n",
         "\t -w password   bind password\n",
         "\t -W            prompt for bind password\n",
         "\t -A            bind without any password (default if no password and DN given)\n",
         "\t -Z            start TLS request\n",
         "\t --help        display this help end exit\n";
   exit(0);
   }
}

use Unicode::MapUTF8 qw(to_utf8 from_utf8 utf8_supported_charset);
use Getopt::Std qw(getopts);
use Net::LDAP;
use Net::LDAP::Util qw( ldap_error_name ldap_error_text) ;

$configfile = "ldap.conf";
$charset    = "ISO-8859-2";

#####################################
##
## read data from ldap.conf file
##

if (-e $configfile){

   $res = read_config();
   if (!$res) {
       print  "An error was occured opening $configfile\n",       
              "Continue without reading this file? (y/n) \[n\]";
       chomp($key=<STDIN>);
       $key="n" if !$key;
       if ($key !~ /^(y|Y)$/) {exit 0;}
   }
}

######################################
##
## initialize variables 
##

$host	     = $conf{"HOST"          } || "localhost";
$port	     = $conf{"PORT"          } || 389;
$basedn	     = $conf{"BASEDN"        } || "";
$binddn      = $conf{"BINDDN"        } || "";
$version     = $conf{"VERSION"       } || 3;
$password    = $conf{"PASSWORD"      } || "";
$verify      = $conf{"TLS_VERIFY"    } || "none";
$sslversion  = $conf{"TLS_VERSION"   } || "tlsv1";
$clientcert  = $conf{"TLS_CLIENTCERT"} || "";
$clientkey   = $conf{"TLS_CLIENTKEY" } || "";
$cafile      = $conf{"TLS_CAFILE"    } || "";
$capath      = $conf{"TLS_CAPATH"    } || "";
$ciphers     = $conf{"TLS_CIPHERS"   } || "";



$file       = "";
$continue   = 0;
$del_subtree= 0;
$anonymous  = 0;
$promptpass = 0;
$ask        = 0;
$start_tls  = 0;


######################################
##
## get parameters from command line
##

@args=@ARGV;

%opt=();
getopts('Ab:cd:f:h:ip:Ww:v:yZ', \%opt) or die "Use --help to check usage\n";

if ($opt{h}) {$host      = $opt{h};shift @args;shift @args;}
if ($opt{p}) {$port      = $opt{p};shift @args;shift @args;}
if ($opt{b}) {$basedn    = $opt{b};shift @args;shift @args;}
if ($opt{d}) {$binddn    = $opt{d};shift @args;shift @args;}
if ($opt{w}) {$password  = $opt{w};shift @args;shift @args;}
if ($opt{v}) {$version   = $opt{v};shift @args;shift @args;} 

if ($opt{f}){
   if (-e $opt{f}){
      $file = $opt{f};
      }
   else {
      print "$opt{f}: No such file or directory\n";
      exit (0);
      }
shift @args;shift @args;
   }

if ($opt{W}){
    $promptpass=1;
    shift @args;
}

if ($opt{c}){
   $continue = 1;
   shift @args;
   }

if ($opt{y}){
   $del_subtree = 1;
   shift @args;
   }

if ($opt{A}) { 
   $anonymous = 1;
   shift @args;
   }

if ($opt{i}) {
   $ask = 1;
   shift @args;
   }

if ($opt{Z}) {
   $start_tls = 1;
   shift @args;
   }

@entries=();

# take DNs for deleting from command line
foreach $arg (@args){
      push @entries, $arg;
}

if ($promptpass){
    system "stty -echo";
    print "Password:";
    chomp($password=<STDIN>);
    print "\n";
    system "stty echo";
}


if ($binddn) {
   $binddn = to_utf8 ({ -string => $binddn, -charset => $charset });
   }

$ldap = Net::LDAP->new( $host, port => $port, version => $version ) or die "$@\n";

if ($start_tls){

   @tls_args = ();
   if ($verify && ($verify !~ /^none$/i)){

      if ($cafile) {
         push @tls_args, 'verify' => $verify, 'cafile' => $cafile;
         }
      elsif ($capath){
         push @tls_args, 'verify' => $verify, 'capath' => $capath;
         }
      else {
         print "Error: You must specify either TLS_CAFILE or TLS_CAPATH, ",
               "if you want to verify server's certificate\n";
         exit 0;
        }
      }

   if ($clientcert){
      if ($clientkey) {
         push @tls_args, 'clientcert' => $clientcert, 'clientkey' => $clientkey;
         }
      else {
         print "Error: You must specify TLS_CLIENTKEY\n";
         exit 0;
         }
      }

   if ($ciphers){
      push @tls_args, 'ciphers' => $ciphers;
      }

   if ($sslversion){
      push @tls_args, 'sslversion' => $sslversion;
      }

   $res = $ldap->start_tls(@tls_args);

   if ($res->code){
      &print_error("start_tls", $res, 1);
      exit 0;
     }
}

if ($anonymous || !($binddn || $password)){

    # anonymous binding
    $res = $ldap->bind;

    if ($res->code) {
       &bind_error($res); 
       exit 0; 
       }
}
else { 

   # authenticate binding 
   $res = $ldap->bind (dn => $binddn, password => $password); 

   if ($res->code) {
       &print_error("bind",$res);
       exit 0;
       }
   }

if (@entries && !$file){
   
   chomp ($entries[$#entries]);
   foreach $dn (@entries){
      
      # remove 'dn:' phrase from dn
      $dn =~ s/^\s*(dn:)\s+//;

      $dn = to_utf8({ -string => $dn, -charset =>$charset });
      $dn = "$dn";
      $first = 1;
      &delete_subtree($dn);
   }
}  

else{

   $stdin=0;
   $fh="FH";

   if (!$file){
      $file  = "-";  
      $stdin = 1; 
      }
   
   open ($fh,"$file") or die "Unable to read from $file\n"; 

   $ldif="";
   while ($line = <$fh>){ 

      chomp($line);
     
      next if ($line =~ /^(version:|\#)/);
    
      if ($ldif && $line =~ /^ (\s*\S+\s*)/){

         # this is continuation of the previous line
         $line =~ s/^ //;
         $ldif .= $line;
         next;
         }
       
      # check if this is the separator of entries
      if ( (!$line) || ($line =~ /^(-?)\s*$/) ){

        # read next if no entry
        next unless $ldif;
 
        # remove white chars, quoting or 'dn:' phrase from dn
        $ldif =~ s/^\s*("|')*(dn:)*\s*("|')*//;
        $ldif =~ s/("|')*$//;

        $dn = to_utf8({ -string => $ldif, -charset =>$charset });
        $dn = "$dn";
        $first = 1;
        $ldif = "";
        $res = &delete_subtree($dn);
        if ($res == 0 ){
           if ($stdin){
              print "Deleted entry $line\n";
              }
           next;
        }
        elsif ($res == 1) {
           &ldif_error();
          }
     }

     $ldif .= $line;

    }

  # delete last entry if it doesn't end with separator
  if ($ldif){

      $ldif =~ s/^\s*("|')*(dn:)*\s*("|')*//;
      $ldif =~ s/("|')*$//;

      $dn = to_utf8({ -string => $ldif, -charset =>$charset });
      $dn = "$dn";
      $first = 1;
      $res = &delete_subtree($dn);
      if ($res == 0 ){
           if ($stdin){
              print "Deleted entry $line\n";
              }
           next;
        }
      }

  close ($fh);
  }
$ldap->unbind;


####################
## Procedures
#####################

 sub delete_subtree{
   my $dn = "$_[0]";
   my $dn_iso = from_utf8({ -string => $dn, -charset=>$charset });
   my $has_sub = 0;
   my $do_delete = 0; 
  
   # searching immediate children of the entry
   $mesg = $ldap->search(
                        base   => $dn,
                        scope  => 'one',
                        filter => '(objectclass=*)'
                        );
   if ($mesg->code) {
      print "\nDeleting entry \"$dn_iso\"\n" unless $stdin; 
      &print_error("delete",$mesg);
      return 1;
      }  
 
   if ($mesg->count == 0 ) {
   
       # there is no children
       $do_delete = 1;
 
       if ($ask){
          do { 
             print "Delete entry $dn_iso? (y/n)[y]";
             chomp($char=<STDIN>);
             $char="y" unless $char;
             if ($char =~ /^n$/i){ 
                $do_delete = 0;
             }
          }while ($char !~ /^(y|n)$/i);
       }        
       if ($do_delete){
           $res = $ldap->delete($dn);
           if ($res->code) {
               &print_error("delete",$res);
               return 1 ;
           }
           else {
               return 0;
           }
       }
       else { 

             # don't want to delete this entry
             return 2;
          }
       }
       else {

            if ($first) {
               if(!$del_subtree){
                    do {
                       if (!$stdin){
                            print "Entry $dn_iso isn't a leaf.\n",
                                  "Delete whole subtree? (y/n)[y]";
                       }
                       else {
                          print "This DN isn't a leaf.\nDelete whole subtree?(y/n)[y]";
                       }
                       chomp($char=<STDIN>);
                       $char="y" unless $char;
                       if ($char =~ /^n$/i) {
                            $do_delete = 0;
                             return 1;
                       }
                    } while ($char !~ /^(y|n)$/i);
               }  
               $first = 0;
            } 
 
            # there are subtree, need to delete each children first
            $has_sub = 0;
            foreach $entry ($mesg->entries){
                $res = &delete_subtree($entry->dn);
                if ($res == 2) {

                   # there is leaving a child  in subtree
                   # cannot delete this entry
                   $has_sub=1;
                   } 
            }


            if (!$has_sub){
                
               # delete only dn without subtree 
               $do_delete=1;
               my $dn_iso=from_utf8({ -string => $dn, -charset=>$charset });
               if ($ask){ 
                  do {
                     print "Delete entry $dn_iso? (y/n)[y]";
                     chomp($char=<STDIN>);
                     $char="y" unless $char;
                     if ($char =~ /^n$/i) {
                        $do_delete=0;
                        }
                     }while ($char !~ /^(y|n)$/i);
                }
               if ($do_delete){    
                  $res = $ldap->delete($dn);     
                  if ($res->code){
                     &print_error("delete",$res);
                     return 1;
                     }
               else {
                   return 0;
                   }
               }
             
            else {
                # means, that this dn has children and cannot be deleted 
                return 2;
                }
            }
          return 2;
          }
}

# read data from ldif.conf file and set approraite variables' values
sub read_config{

   %conf = ();
   open (CONF, "$configfile") or return 0;
   while (<CONF>){
      chomp;
      next if (/\#/ or /^\s+$/);
      (/(.*?)\s+(.*?)\s*$/) && ($conf{$1}=$2);
      }
   close (CONF);
}

# set bahavior of program in case of occurring an error
sub ldif_error{

   if ($continue || $stdin){
      $ldif = "";
      next;
      }
   else {
      $ldif = "";
      last;
      }
}

sub print_error{

   my ($text, $res) = @_;
       print "ldap_$text: Error ",$res->code, ": ", $res->error ,"\n";
       print ldap_error_text($res->code);
}

