#!/usr/bin/perl

BEGIN {
if ($ARGV[0] eq "--help"){
   print "Modify entries from an LDAP server\n\n",
         "Usage:  modify.pl [options]\n",
         "        Entries are read from stdin or from the file\n",
         "        specified by \"-f file\"\n",
         "Options:\n",
         "\t -f file       read operation from \"file\"\n",
         "\t -c            continue modifing next entries after reporting an error\n",
         "\t -h host       LDAP serwer (default: localhost)\n",
         "\t -p port       port on LDAP server (default: 389)\n",
         "\t -v version    protocol version to be used (default: 3)\n",
         "\t -d binddn     bind DN\n",
         "\t -w password   bind password\n",
         "\t -W            prompt for bind password\n",
         "\t -A            bind without any password (default if no password and DN given)\n",
         "\t -Z            start TLS request\n",
         "\t --help        display this help end exit\n";

   exit(0);
   }
}

use Unicode::MapUTF8 qw(to_utf8 from_utf8 utf8_supported_charset);
use Getopt::Std qw(getopts);
use Net::LDAP;
use Net::LDAP::Util qw( ldap_error_name ldap_error_text) ;

$configfile = "ldap.conf";
$charset    = "ISO-8859-2";

#####################################
##
## read data from ldap.conf file
##

if (-e $configfile){

   $res = read_config();
   if (!$res) {
       print  "An error was occured opening $configfile\n",       
              "Continue without reading this file? (y/n) \[n\]";
       chomp($key=<STDIN>);
       $key="n" if !$key;
       if ($key !~ /^(y|Y)$/) {exit 0;}
   }
}

######################################
##
## initialize variables 
##

$host	     = $conf{"HOST"          } || "localhost";
$port	     = $conf{"PORT"          } || 389;
$basedn	     = $conf{"BASEDN"        } || "";
$binddn      = $conf{"BINDDN"        } || "";
$version     = $conf{"VERSION"       } || 3;
$password    = $conf{"PASSWORD"      } || "";
$verify      = $conf{"TLS_VERIFY"    } || "none";
$sslversion  = $conf{"TLS_VERSION"   } || "tlsv1";
$clientcert  = $conf{"TLS_CLIENTCERT"} || "";
$clientkey   = $conf{"TLS_CLIENTKEY" } || "";
$cafile      = $conf{"TLS_CAFILE"    } || "";
$capath      = $conf{"TLS_CAPATH"    } || "";
$ciphers     = $conf{"TLS_CIPHERS"   } || "";


$file       = "";
$continue   = 0;
$anonymous  = 0;
$promptpass = 0;
$start_tls  = 0;


######################################
##
## get parameters from command line
##

%opt=();
getopts('Ab:cd:f:h:p:Ww:v:Z', \%opt) or die "Use --help option to check syntax\n";

if ($opt{h}) {$host      = $opt{h};}
if ($opt{p}) {$port      = $opt{p};}
if ($opt{b}) {$basedn    = $opt{b};}
if ($opt{d}) {$binddn    = $opt{d};}
if ($opt{w}) {$password  = $opt{w};}
if ($opt{v}) {$version   = $opt{v};}
 
if ($opt{f}){
   if (-e $opt{f}){
      $file = $opt{f};
      }
   else {
      print "$opt{f}: No such file or directory\n";
      exit (0);
      }
   }

if ($opt{W}){
    $promptpass=1;
    }

if ($opt{c}){
   $continue = 1;
   }

if ($opt{A}) { 
   $anonymous = 1;
   }

if ($opt{Z}) {
   $start_tls = 1;
   }

if ($promptpass){
    system "stty -echo";
    print "Password:";
    chomp($password=<STDIN>);
    print "\n";
    system "stty echo";
}

if ($binddn) {
   $binddn = to_utf8 ({ -string => $binddn, -charset => $charset });
   }

$ldap = Net::LDAP->new( $host, port => $port, version => $version ) or die "$@\n";

if ($start_tls){

   @tls_args = ();
   if ($verify && ($verify !~ /^none$/i)){

      if ($cafile) {
         push @tls_args, 'verify' => $verify, 'cafile' => $cafile;
         }
      elsif ($capath){
         push @tls_args, 'verify' => $verify, 'capath' => $capath;
         }
      else {
         print "Error: You must specify either TLS_CAFILE or TLS_CAPATH, ",
               "if you want to verify server's certificate\n";
         exit 0;
        }
      }

   if ($clientcert){
      if ($clientkey) {
         push @tls_args, 'clientcert' => $clientcert, 'clientkey' => $clientkey;
         }
      else {
         print "Error: You must specify TLS_CLIENTKEY\n";
         exit 0;
         }
      }

   if ($ciphers){
      push @tls_args, 'ciphers' => $ciphers;
      }

   if ($sslversion){
      push @tls_args, 'sslversion' => $sslversion;
      }

   $res = $ldap->start_tls(@tls_args);

   if ($res->code){
      &print_error("start_tls", $res, 1);
      exit 0;
     }
}

if ($anonymous || !($binddn || $password)){

    # anonymous binding
    $res=$ldap->bind;

    if ($res->code) {
       &print_error("bind",$res,1);
       exit 0;
       }
}
else { 

   # authenticate binding 
   $res = $ldap->bind (dn => $binddn, password => $password); 

   if ($res->code) {
       &print_error("bind",$res,1);
       exit 0;
       }
   }

$fh="FH";
$stdin=0;

if (!$file){
   $file  = "-";
   $stdin = 1;
   }

open ($fh,"$file") or die "Unable to read from $file\n"; 

@ldif=();
$changetype  = "";

while ($line=<$fh>){
  
   chomp($line);
   if (@ldif && $line =~ /^ (\s*\S+\s*)/){
      
      # this is continuation of the previous line
      $last = scalar @ldif;
      $line =~ s/^ //;
      $ldif[$last-1].=$line;
      next;
      }

   next if ($line =~ /^(version:|\#)/);

   # check if this is the separator of entries
   if ( (!$line) || ($line =~ /^(-?)\s*$/) ){
         
      # read next if no entry
      next unless @ldif;   
         
      if ($changetype eq "modify") {

         # modifed entry can have "-" as separator
         if ($line =~ /^-\s*$/){
            push (@ldif,$line);
            next;
            }
         else {
             
           # there aren't next operations, do modify ...    
            if ( &modify_entry(\@ldif) == 0){ 
               @ldif=();
               $changetype  = "";  
               next;
               }
            else {
               &ldif_error();
               }
            } 
         }
      
      if ($changetype eq "delete"){
            
         if (&delete_entry(\@ldif) == 0){
            @ldif=();
            $changetype  = "";
            next;
            }
         else {
             &ldif_error();
             }
         }
      if ($changetype eq "add"){

         if (&add_entry(\@ldif) == 0){
            @ldif=();
            $changetype  = "";
            next;
            }
          else {
              &ldif_error();
             }
         }      
 
       if (($changetype eq "moddn") || ($changetype eq "modrdn")){

         $res = &moddn_entry(\@ldif);
         if ($res == 0){
            @ldif=();
            $changetype  = "";
            next;
            }
         elsif ($res == 2){
            print "Renaming/Moving entry $ldif[0]\n" unless $stdin;
            print "Error: one of the parameters was specified more than once\n\n";
            &ldif_error();
            }
          else {
            &ldif_error(); 
            }
         }
   } 

             
   # check if this is the first line of entry
   if (!@ldif){
                    
      #  first line of entry must begin with 'dn:'
      if ($line !~ s/^dn:\s+//){
          print "Error: invalid format entry \"$line\"\n", 
                "   (first line of entry must begin with \'dn:\')\n\n";
 
          &ldif_error();
          }
      
    }    
  
    # isn't first line, check format next lines of entry
    if (@ldif == 1 && (!$changetype)){
 
       if ($line !~ /^changetype:\s+(add|delete|modify|moddn|modrdn)\s*$/){
          print "Error: invalid format entry \"$ldif[0]\"\n",
                " line: $line (expect valid changetype)\n\n";     
          &ldif_error(); 
          }
       else {
           $changetype=$1;
           next;
           }
       }

     # add line to an array containing the lines of entry
     push @ldif, $line;
}

# modify last entry in case not ending with separator
if (@ldif){
   
   if ($changetype eq "modify") {
      &modify_entry(\@ldif);
      }
   if ($changetype eq "delete"){
      &delete_entry(\@ldif);
      }

   if ($changetype eq "add"){
      &add_entry(\@ldif);
      }

   if (($changetype eq "moddn") || ($changetype eq "modrdn")){
      $res = &moddn_entry(\@ldif);
      if ($res == 2){
         print "Error: one of the parameters was specified more than once\n\n"
         }
      }
}
close ($fh);

$ldap->unbind;
exit 0;

####################
## Procedures
#####################

sub add_entry{

   my @ldif = @{$_[0]};
   my @data = ();

   # retrive 'dn' from @ldif array
   my $dn_iso = $ldif[0];
   my $dn     = to_utf8 ({ -string => $dn_iso, -charset =>$charset });

   # remove 'dn' from @ldif array 
   shift @ldif;
  
   # check if there are attibutes of the entry
   if (!@ldif) {
      print "Error: no attributes to add\n\n";  
      return 1;
      }
  
   
   # get all attributes of the entry and preprare it for add method 
   while(@ldif){
 
      $line = shift @ldif;    

      if ($line =~ /^([-;\w]+):(:?|<?)\s+(.*)$/){
             
         $attr = $1;
         $val  = $3;

         if ($2 eq ":") {
             eval { require MIME::Base64 };

             if ($@) {
                print "Modifying entry \"$dn_iso\"\n" unless $stdin;
                print "Error: unable to load MIME::Base64 package\n\n";
                return 1;
                }
            $val = MIME::Base64::decode($val);
            }

         elsif ($2 eq "<") {
             if ($val =~ /^file:\/\/(.*)/){
                my $file = $1;
                $old = $/;
                local $/ = undef;
                $res = open ("fh", $file);
                unless ($res) {
                   print "Error: Unable to open $file file\n\n";
                   return 1;
                   }
                $val = <fh>;
                close("fh");
                $/ =  $old;
                }
             else {
                 print "Error: invalid format entry \"$dn_iso\"\n",
                       " line: $line (supported \'file://\' url only)\n\n";
                 return 1;
                 }
             }
         else{
            $val = to_utf8 ({ -string => $val, -charset =>$charset });
            }

         push @data, $attr => $val;
         }
      else {
          print "Error: invalid format entry \"$dn_iso\"\n",
                " line: $line \n\n";
          return 1;
          }
   }
   $data = \@data;
   
   # add the entry to an LDAP server
   $res = $ldap->add ( $dn, attrs => $data );

   # check result and report errors
   if ($res->code){
      print "Adding entry \"$dn_iso\"\n";
      &print_error("add", $res);
      return 2;
      }
   else {
       print "Added an entry \"", $dn_iso, "\"\n";
       print "\n" if $stdin;
       }
   return 0;
}


sub delete_entry{

   my @ldif = @{$_[0]};
   my $dn_iso = $ldif[0];
   my $dn     = to_utf8({ -string => $dn_iso, -charset=>$charset });

   shift @ldif;

   if (@ldif) {

      print "Error: invalid format entry \"$dn_iso\"\n\n";
      return 1;
      }

   $res = $ldap->delete($dn);

   if ($res->code){
      print "Deleting entry $dn_iso\n" unless $stdin;
      &print_error("delete",$res,1);
      return 1;
      }
   else {
      print "Deleted entry \"$dn_iso\"\n";
      print "\n" if $stdin;
      return 0;
      }
}

sub modify_entry{

   my @ldif = @{$_[0]};
   my $dn_iso = $ldif[0];
   my $dn     = to_utf8({ -string => $dn_iso, -charset => $charset });

   shift @ldif;

   my $modifytype  = "";
   my $mustmodtype = 1;
   my @data = ();

   if (!@ldif) {
      print "Error: invalid format entry \"$dn_iso\"\n",
            " no parameters to modify\n\n";
      return 1;
      }

   my @Changes = ();
   my @DD = ();
   my @AA = ();
   my $i = 0;
   my $j = 0;
   
   # read each line of entry
   while (@ldif){

      $line = shift @ldif;

      if ($mustmodtype){

         if ($line =~ /^(add|delete|replace):\s+([-;\w]+)$/) {
            $mustmodtype = 0;
            $modifytype  = $1;
            $attr = $2;
            next;
            }
         else {
             print "Error: invalid format entry \"$dn_iso\"\n",
                   " line: $line (expect valid modifytype and attributename)\n\n";
             return 1;
             }
      }
      elsif ($line =~ /^($attr):(:?|<?)\s+(.*)$/i){
                                 
          $val = $3;

          if ($2 eq ":") {
             eval { require MIME::Base64 };

             if ($@) {
                print "Modifying entry \"$dn_iso\"\n" unless $stdin;
                print "Error: unable to load MIME::Base64 package\n\n";
                return 1;
                }
            $val = MIME::Base64::decode($val);
            }
          elsif ($2 eq "<") {
             if ($val =~ /^file:\/\/(.*)/){
                my $file = $1;
                $old = $/;
                local $/ = undef;
                $res = open ("fh", $file); 
                unless ($res) {
                   print "Error: Unable to open $file file\n\n";
                   return 1;
                   }
                $val = <fh>;
                close("fh");
                $/ =  $old;
                } 
             else {
                  print "Error: invalid format entry \"$dn_iso\"\n", 
                        " line: $line (supported \'file://\' url only)\n\n";
                  return 1;
                  }
             }
          else{
             $val = to_utf8 ({ -string => $val, -charset =>$charset });
             }
          push @data, $val;
          next;
          }
      
      elsif ($line =~ /^-\s*$/){

         if (@data){

            $DD[$i] = [@data];
            @tmp = ();
            push @tmp, $attr=>$DD[$i];
            
            $AA[$i] = [@tmp];
            push @Changes, $modifytype=>$AA[$i];
            $i++;
            }

         elsif ($modifytype eq "delete"){
           
             $DD[$i] = [()];
             @tmp = ();
             push @tmp, $attr=>$DD[$i];

             $AA[$i] = [@tmp];
             push @Changes, $modifytype=>$AA[$i];
             $i++;

         }
         else {
             print "Error: invalid format entry \"$dn_iso\"\n\n";
             return 1;
         }

         @data = ();
         $modifytype = "";
         $mustmodtype = 1;
         }

      else {
          print "Error: invalid format entry \"$dn_iso\"\n",
                " line: $line\n\n";
          return 1;
          }
      }

   # modifing last attribute of entry
   if (@data || ($modifytype eq "delete")){
       if (@data){
          
          $DD[$i] = [@data];
       }
       else { 

          $DD[$i] = [()];
       } 

       @tmp = ();
       push @tmp, $attr=>$DD[$i];
          
       $AA[$i] = [@tmp];
       push @Changes, $modifytype => $AA[$i];
       $i++;

   }
   else {
       print "Error: invalid format entry \"$dn_iso\"\n\n";
       return 1;
   }


   if (@Changes) {

      $Changes = \@Changes;
      $res  = $ldap->modify($dn, changes => $Changes);


      if ($res->code) {
              print "Modifying entry \"$dn_iso\"\n" unless $stdin;
              &print_error("modify",$res,1);
              return 1;
              }
           else {
              print "Modified entry \"$dn_iso\"\n";
              print "\n" if $stdin;
              return 0;
           }
   }
}

sub moddn_entry{
   
   my @ldif = @{$_[0]};
   my @data = ();

   # retrive 'dn' from @ldif array
   my $dn_iso = $ldif[0];
   my $dn     = to_utf8 ({ -string => $dn_iso, -charset =>$charset });

   # remove 'dn' from @ldif array
   shift @ldif;

   if (!@ldif){
       print "Error: invalid format entry \"$dn_iso\"\n",
             " (not given parameters)\n\n";
       return 1;
      }
   
   my %Moddn=();
   while (@ldif){

      $line = shift @ldif;
 
      if ($line =~ /^newrdn:\s+(.*)$/){

         if (exists ($Moddn{'newrdn'})){
            return 2; 
            }

         $Moddn{'newrdn'} = $1; 
         $newrdn = to_utf8 ({ -string => $Moddn{'newrdn'}, -charset => $charset });  
         push (@data, 'newrdn' => $newrdn);    
        }

        elsif ($line =~ /^deleteoldrdn:\s+(0|1)$/){

          if (exists ($Moddn{'delete'})){
              return 2;
             }

         $Moddn{'delete'} = $1; 
            if ($1 == 1){
            push (@data, 'deleteoldrdn' => '1');
             }
         }

         elsif ($line =~ /^newsuperior:\s+(.*)$/){

            if (exists ($Moddn{'newsuperior'})){
               return 2;
              }

            $Moddn{'newsuperior'} = $1;
            $newsuperior = to_utf8 ({ -string => $Moddn{'newsuperior'}, -charset => $charset});
            push (@data, 'newsuperior' => $newsuperior);
            }
         else {
            print "Error: invalid format entry \"$dn_iso\"\n",
                  " line: $line\n\n";
            return 1;
            }   
   }

   # check needed parameters and do moddn
   unless (exists($Moddn{'newrdn'}) && exists($Moddn{'delete'})) {
       print "Error: invalid format entry \"$dn_iso\"\n",
             "       (not given 'newrdn' or 'deleteoldrdn' parameter)\n\n";
       return 1;
       }

   $res = $ldap->moddn($dn,@data);

   if ($res->code){
       print "Renaming/Moving entry \"$dn_iso\"\n" unless $stdin;
       &print_error("moddn",$res,1);
       return 1;
       }
   else {
      print "Renamed/Moved entry \"$dn_iso\"\n"; 
      print "\n" if $stdin;
      }
}


sub print_error{

   my ($text, $res, $ext) = @_;
       print "ldap_$text: Error ",$res->code, ": ", $res->error ,"\n";
       print " ", ldap_error_text($res->code) if ($ext == 1);
       print "\n";
}


sub ldif_error{

   if ($continue || $stdin){
      @ldif=();
      $changetype= "";
      next;
      }

   else {
      @ldif=();
      last;
      }
}


sub read_config{

   %conf = ();
   open (CONF, "$configfile") or return 0;

   while (<CONF>){
      chomp;
      next if (/\#/ or /^\s+$/);
      (/(.*?)\s+(.*?)\s*$/) && ($conf{$1} = $2);
      }
   close (CONF);
}
