#!/usr/bin/perl

BEGIN {
if ($ARGV[0] eq "--help"){
   print "Usage:  search.pl [options] [filter [attributes]]\n",
         "\t filter      RFC-2254 compliant LDAP search filter (default: (objectclass=*))\n",
         "\t attributes  whitespace-separated list of attribute names\n",
         "Options:\n",
         "\t -b basedn     base dn for search\n",
         "\t -r deref      one of never (default), search, find,always\n",
         "\t -s scope      one of base, one, or sub (default) (search scope) )\n",
         "\t -L            print responses in LDIF format\n",
         "\t -n            retrieve attribute names only\n",
         "\t -t timelimit  time limit (in seconds) for search (default: 0 means no limit)\n",
         "\t -z sizelimit  size limit (in entries) for search (default: 0 means no limit)\n",
         "\t -h host       LDAP serwer (default: localhost)\n",
         "\t -p port       port on LDAP server (default: 389)\n", 
         "\t -v version    protocol version to be used (default: 3)\n",
         "\t -d binddn     bind DN\n",
         "\t -w password   bind password\n",
   	 "\t -W            prompt for bind password\n",
         "\t -A            bind without any password (default if no password and DN given)\n",
         "\t -Z            start TLS request\n",
    	 "\t --help        display this help end exit\n";
   exit(0);
   }
}

use Unicode::MapUTF8 qw(to_utf8 from_utf8 utf8_supported_charset );
use Getopt::Std;
use Net::LDAP;
use Net::LDAP::Util qw( ldap_error_name ldap_error_text) ;

$configfile="ldap.conf";
$charset="ISO-8859-2";

#####################################
##
## get data from ldap.conf file
##

if (-e $configfile){
   $res = read_config();
   if (!$res) {
       print  "An error has occured opening $configfile\n",
              "Continue without reading this file? (y/n) \[n\]";
       chomp($key=<STDIN>);
       $key="n" if !$key;
       if ($key !~ /^(y|Y)$/) {exit 0};
   }
}

######################################
##
## initialize variables 
##

$host	     = $conf{"HOST"          } || "localhost";
$port	     = $conf{"PORT"          } || 389;
$basedn	     = $conf{"BASEDN"        } || "";
$binddn      = $conf{"BINDDN"        } || "";
$version     = $conf{"VERSION"       } || 3;
$password    = $conf{"PASSWORD"      } || "";
$scope       = $conf{"SCOPE"         } || "sub";
$deref	     = $conf{"DEREF"         } || "never";
$timelimit   = $conf{"TIMELIMIT"     } || 0;
$sizelimit   = $conf{"SIZELIMIT"     } || 0;
$verify      = $conf{"TLS_VERIFY"    } || "";
$sslversion  = $conf{"TLS_VERSION"   } || "tlsv1";
$clientcert  = $conf{"TLS_CLIENTCERT"} || "";
$clientkey   = $conf{"TLS_CLIENTKEY" } || "";
$cafile      = $conf{"TLS_CAFILE"    } || "";
$capath      = $conf{"TLS_CAPATH"    } || "";
$ciphers     = $conf{"TLS_CIPHERS"   } || "";


$filter        = "(objectclass=*)";
$attrs	       = qw(*);
$anonymous     = 0;
$promptpass    = 0;
$display_ldif  = 0;
$wrap          = 76;
$objectClass   = 1;
$att_names     = 0;
$start_tls     = 0;

######################################
##
## get parameters from command line
##

@args = @ARGV;

%opt=();
getopts('Ab:d:Lns:h:p:r:Ww:v:t:Zz:', \%opt) or die "Use --help option to check syntax\n";

if ($opt{h}) {$host      = $opt{h};shift @args;shift @args;}
if ($opt{p}) {$port      = $opt{p};shift @args;shift @args;}
if ($opt{b}) {$basedn    = $opt{b};shift @args;shift @args;}
if ($opt{s}) {$scope     = $opt{s};shift @args;shift @args;}
if ($opt{d}) {$binddn    = $opt{d};shift @args;shift @args;}
if ($opt{r}) {$deref     = $opt{r};shift @args;shift @args;}
if ($opt{w}) {$password  = $opt{w};shift @args;shift @args;}
if ($opt{v}) {$version  =  $opt{v};shift @args;shift @args;}

if ($opt{t} || ($opt{t} eq "0")) {
              $timelimit = $opt{t};shift @args;shift @args;}

if ($opt{z} || ($opt{z} eq "0")) {
              $sizelimit = $opt{z};shift @args;shift @args;}


if ($opt{W}){
   $promptpass = 1;
   shift @args;
   }

if ($opt{A}) {
   $anonymous = 1;
   shift @args;
   }

if ($opt{L}) {
   $display_ldif = 1;
   shift @args;
   }

if ($opt{n}) {
   $att_names = 1;
   shift @args;
   }

if ($opt{Z}) {
   $start_tls = 1;
   shift @args;
   }

if ($promptpass){
    system "stty -echo";
    print "Password:";
    chomp($password=<STDIN>);
    print "\n";
    system "stty echo";
}

@attrs = ();
$filter_arg = 0;

# get filter and attributes parameters
while (@args){
   $arg = shift @args;

   if ($arg !~ /^[\w\+\*\(]/){
      print "Uknown option $arg or invalid parameter\n",
            "Use --help to check usage\n";
      exit 0;
      }

   else {

      if (!$filter_arg){
         $filter = $arg;
         $filter_arg = 1;
         }
      else {
       push @attrs, $arg;   

       # set this if want to display objectClass attribute
       $objectClass = 2 if ($arg =~ /objectclass/i);  
       } 
   }      
}

# take attributes array as reference for search method 
if (@attrs){
   $attrs = \@attrs;

# won't display objectClass attribute
   $objectClass = 0 unless ($objectClass == 2);
   }
     
$filter_iso = $filter;
$filter = to_utf8 ({ -string => $filter_iso, -charset => $charset });

if ($binddn){
   $binddn = to_utf8 ({ -string => $binddn, -charset => $charset });
   }

if ($basedn){
   $basedn = to_utf8 ({ -string => $basedn, -charset => $charset });
   }

$ldap = Net::LDAP->new($host, port => $port, version => $version ) or die "$@\n";

if ($start_tls){

   @tls_args = (); 
   if ($verify && ($verify !~ /^none$/i)){

      if ($cafile) {
         push @tls_args, 'verify' => $verify, 'cafile' => $cafile;
         }
      elsif ($capath){
         push @tls_args, 'verify' => $verify, 'capath' => $capath;  
         }
      else {
         print "Error: You must specify either TLS_CAFILE or TLS_CAPATH, ",
               "if you want to verify server's certificate\n";
         exit 0;
        }
      }

   if ($clientcert){
      if ($clientkey) {
         push @tls_args, 'clientcert' => $clientcert, 'clientkey' => $clientkey;
         }
      else {
         print "Error: You must specify TLS_CLIENTKEY\n";
         exit 0;
         }
      }

   if ($ciphers){
      push @tls_args, 'ciphers' => $ciphers;
      }

   if ($sslversion){
      push @tls_args, 'sslversion' => $sslversion;
      }
 
   $res = $ldap->start_tls(@tls_args);

   if ($res->code){
      &print_error("start_tls", $res, 1);
      exit 0;
     }
}

if ($anonymous || !($binddn || $password)){
   
    # anonymous binding 
    $res = $ldap->bind;

    if ($res->code){
       &print_error("bind",$res,1);
       exit 0;
       }
    }
else{

    # authenticate binding
    $res = $ldap->bind (dn => $binddn, password => $password); 
  
    if ($res->code){
       &print_error("bind",$res,1);
       exit 0;
       }
    }

$mesg = $ldap->search ( 
               	base	  => $basedn,
             	scope	  => $scope,
		filter	  => $filter,
		attrs	  => $attrs,
                deref 	  => $deref,
                sizelimit => $sizelimit, 
                timelimit => $timelimit 
                );
 
if ($mesg->code) {
   &print_error("search",$mesg,1);
   exit(0);
   }

if ($display_ldif){
   print "version: 1\n\n";

   foreach $entry ($mesg->entries){

      &display_entry_ldif($entry);
      print "\n";
      }

   }
else {
  
    print "\nfilter:    $filter_iso\n";

    # print comma-separated list od requested attributes
    if (@attrs) {
        print "attributes: ", join(",", @attrs) ,"\n\n";
        } 
    else {
        print "attributes: all\n\n";
        }
   
     foreach $entry ($mesg->entries){

        &display_entry($entry);
        print "\n";
        }
    print "Search result: ",$mesg->count, " entries\n\n" unless $opt{L};
    }

$ldap->unbind;

####################
## Procedures
#####################

sub display_entry_ldif{
    
    my $entry = $_[0];
    my $line = "";

    $dn = from_utf8({ -string => $entry->dn, -charset => $charset });
    $line = "dn: ".$dn;
    print &wrap($line, $wrap), "\n";

    # display objectClass attribute first if requested
    if ($objectClass){
       $att = "objectClass";
       @val = $entry->get_value($att);

       foreach $v (@val){
           print "$att:";
           print " $v" unless $att_names;
           print "\n";
           }
    }     
       
    # get list of attributes
    @attrs = sort $entry->attributes;      
 
    foreach $attr ( @attrs ){

         # objectClass attribute is already displayed
         next if ($attr =~ /objectclass/i);

         # get list of values of attr attribute
         @val = $entry->get_value($attr);

         foreach $v (@val){
            
            if ($att_names) {
               print "$attr:\n";
               }
            else {
               $v_iso = from_utf8({ -string => $v, -charset => $charset });

               # check if value contain binary data
               if (&binary(\$v_iso)){

                   # binary data, load package and encode 
                   eval { require MIME::Base64 };
                   if ($@) {
                      print "Error: unable to load MIME::Base64 package\n\n";
                      return 1;
                      }
                   $line = "$attr" . ":: ". MIME::Base64::encode($v,"");
                   }
                 else{

                     # only Latin2 characters, can display it                      
                     $line = "$attr: $v_iso";
                     }
                # folded into lines if lenght of lines is more as $wrap
                print &wrap($line, $wrap), "\n";
                }   
            }
         }
}

sub display_entry{
    
     my $entry = $_[0];
     $dn = from_utf8({ -string => $entry->dn, -charset => $charset });
     @arg = split "," ,$dn; 
     print "# ";
     
     # print name of entry as comment
     for ($i=0;$i<@arg;$i++){
         ($a,$v) = split "=", $arg[$i];
         print "$v";
         if ($i != @arg-1){
            print ", ";
            }
         }
     
     print "\ndn: ",$dn,"\n";
     $len = 0;

     # get attriutes names
     @attrs = sort $entry->attributes;   
     
     # check lenght of longest attribute name
     if (!$att_names){
     
        foreach $attr ( @attrs ){
            $length = length $attr;
            $len = $length if $length > $len;
            }

        # define space  
        $spc="\n  " . " " x $len;
        }
      
      # print objectClass attribute first if requested
      if ($objectClass){
         $attr="objectClass";
         if ($att_names){
               print "$attr:\n";
               }
         else {
           
            # get list of values of objectClass attribute
            @val = $entry->get_value($attr);
            printf "%${len}s: ",$attr;
            $i=0;
     
            # print values
            foreach $v (@val){
                print $spc if $i++;
                print "$v";
                }
            print "\n";
            }
       }
     
     # print remaining attributes and values 
     foreach $attr ( @attrs ){
     next if ($attr =~ /objectclass/i);  
         if ($att_names){
            print "$attr:\n";
            } 
         else{
            @val = $entry->get_value($attr);
            printf "%${len}s: ",$attr;
            $i=0;
            
            foreach $v (@val){
                $v = from_utf8({ -string => $v, -charset => $charset }); 
               
                # check if values contain binary data
                if (&binary(\$v)){
                   
                   # won't to display binary data
                   $v = "binary data";
                   }
                print $spc if $i++;
                print "$v";
                }
            print "\n";
            }
      }
      
}

# read data from config file
sub read_config{

     %conf=();
     open (CONF, "$configfile") or return 0;
     while (<CONF>){
           chomp;
           next if (/\#/ or /^\s+$/);
          (/(.*?)\s+(.*?)\s*$/) && ($conf{$1}=$2);
          }
      close (CONF);
}


sub print_error{

   my ($text, $res) = @_;
       print "ldap_$text: Error ",$res->code, ": ", $res->error ,"\n";
       print ldap_error_text($res->code);
}

sub wrap {

    my $pos = $_[1];
    while($pos < length($_[0])) {
      substr($_[0],$pos,0) = "\n ";
      $pos += $_[1]+1;
    }
   $_[0];
}

sub binary{

    return 1 unless 
          (${$_[0]} =~ /^[\x20-\x7e                     # ASCII characters
                          \xa1 \xb1 \xc6 \xe6 \xea \xca # polish diacritics
                          \xb3 \xa3 \xf1 \xd1 \xf3 \xd3
                          \xb6 \xa6 \xbc \xac \xbf \xaf
                         ]+$/);
    return 0;
}
